package extras

import (
	"net"
	"net/netip"
	"net/url"
	"reflect"
	"time"
	"unsafe"
	_ "unsafe"

	"codeberg.org/gruf/go-mangler"
)

func init() {
	// Register standard library performant manglers.
	mangler.Register(reflect.TypeOf(net.IPAddr{}), mangle_ipaddr)
	mangler.Register(reflect.TypeOf(&net.IPAddr{}), mangle_ipaddr_ptr)
	mangler.Register(reflect.TypeOf(netip.Addr{}), mangle_addr)
	mangler.Register(reflect.TypeOf(&netip.Addr{}), mangle_addr_ptr)
	mangler.Register(reflect.TypeOf(netip.AddrPort{}), mangle_addrport)
	mangler.Register(reflect.TypeOf(&netip.AddrPort{}), mangle_addrport_ptr)
	mangler.Register(reflect.TypeOf(time.Time{}), mangle_time)
	mangler.Register(reflect.TypeOf(&time.Time{}), mangle_time_ptr)
	mangler.Register(reflect.TypeOf(url.URL{}), mangle_url)
	mangler.Register(reflect.TypeOf(&url.URL{}), mangle_url_ptr)
}

//go:linkname eface_data codeberg.org/gruf/go-mangler.eface_data
func eface_data(a any) unsafe.Pointer

//go:linkname append_uint32 codeberg.org/gruf/go-mangler.append_uint32
func append_uint32([]byte, uint32) []byte

//go:linkname append_uint64 codeberg.org/gruf/go-mangler.append_uint64
func append_uint64([]byte, uint64) []byte

//go:linkname time_sec time.(*Time).sec
func time_sec(*time.Time) int64

//go:linkname time_nsec time.(*Time).nsec
func time_nsec(*time.Time) int32

//go:linkname time_stripMono time.(*Time).stripMono
func time_stripMono(*time.Time)

func mangle_url(buf []byte, a any) []byte {
	u := (*url.URL)(eface_data(a))
	return append(buf, u.String()...)
}

func mangle_url_ptr(buf []byte, a any) []byte {
	if ptr := (*url.URL)(eface_data(a)); ptr != nil {
		s := ptr.String()
		buf = append(buf, '1')
		return append(buf, s...)
	}
	buf = append(buf, '0')
	return buf
}

func mangle_time(buf []byte, a any) []byte {
	t := *(*time.Time)(eface_data(a))
	time_stripMono(&t) // force non-monotonic time value.
	buf = append_uint64(buf, uint64(time_sec(&t)))
	buf = append_uint32(buf, uint32(time_nsec(&t)))
	return buf
}

func mangle_time_ptr(buf []byte, a any) []byte {
	if ptr := (*time.Time)(eface_data(a)); ptr != nil {
		t := *ptr
		buf = append(buf, '1')
		time_stripMono(&t) // force non-monotonic time value.
		buf = append_uint64(buf, uint64(time_sec(&t)))
		buf = append_uint32(buf, uint32(time_nsec(&t)))
		return buf
	}
	buf = append(buf, '0')
	return buf
}

func mangle_ipaddr(buf []byte, a any) []byte {
	i := *(*net.IPAddr)(eface_data(a))
	buf = append(buf, i.IP...)
	buf = append(buf, i.Zone...)
	return buf
}

func mangle_ipaddr_ptr(buf []byte, a any) []byte {
	if ptr := (*net.IPAddr)(eface_data(a)); ptr != nil {
		buf = append(buf, '1')
		buf = append(buf, ptr.IP...)
		buf = append(buf, ptr.Zone...)
		return buf
	}
	buf = append(buf, '0')
	return buf
}

func mangle_addr(buf []byte, a any) []byte {
	i := (*netip.Addr)(eface_data(a))
	b, _ := i.MarshalBinary()
	return append(buf, b...)
}

func mangle_addr_ptr(buf []byte, a any) []byte {
	if ptr := (*netip.Addr)(eface_data(a)); ptr != nil {
		buf = append(buf, '1')
		b, _ := ptr.MarshalBinary()
		return append(buf, b...)
	}
	buf = append(buf, '0')
	return buf
}

func mangle_addrport(buf []byte, a any) []byte {
	i := (*netip.AddrPort)(eface_data(a))
	b, _ := i.MarshalBinary()
	return append(buf, b...)
}

func mangle_addrport_ptr(buf []byte, a any) []byte {
	if ptr := (*netip.AddrPort)(eface_data(a)); ptr != nil {
		buf = append(buf, '1')
		b, _ := ptr.MarshalBinary()
		return append(buf, b...)
	}
	buf = append(buf, '0')
	return buf
}
